"""
Functions to check/process arguments in topiary functions.
"""
from .paralog_patterns import check_paralog_patterns
from .topiary_dataframe import check_topiary_dataframe
from .standard import check_bool, check_float, check_int, check_iter
